import time
import logging

import numpy as np
import torch
import torch.nn.functional as F
from torch_sparse import SparseTensor
from tqdm import tqdm

from .model import Model, Encoder
from .sampler import NeighborSampler
from .utils import setup_seed, load_data, get_positivePairs, clustering_st

def run(args):
    """
    Main training and inference pipeline.
    """
    load_start = time.time()
    # Set random seed for reproducibility
    setup_seed(args.seed)

    # Select device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    logging.info(f"Using device: {device}")

    # Load data and graph structure
    x, edge_index, y, adata, n_classes, expr = load_data(
        data_path=args.data_path,
        dataset=args.dataset,
        phenoLabels=args.phenoLabels,
        nicheLabels=args.nicheLabels,
        embedding_type=args.embedding_type,
        radius=args.radius,
        k_neighborhood=args.k_neighborhood,
        hvg=args.hvg,
    )
    x = x.to(device)
    if expr is not None:
        expr = expr.to(device)
        
    # Build sparse adjacency tensor
    adj = SparseTensor(
        row=edge_index[0],
        col=edge_index[1],
        value=torch.ones(edge_index.shape[1]),
        sparse_sizes=(x.shape[0], x.shape[0]),
    )

    # Create NeighborSampler loaders
    train_loader = NeighborSampler(
        edge_index,
        adj,
        is_train=True,
        node_idx=None,
        wt=args.wt,
        wl=args.wl,
        p=args.p,
        q=args.q,
        sizes=args.size,
        batch_size=args.batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=8,
    )
    test_loader = NeighborSampler(
        edge_index,
        adj,
        is_train=False,
        node_idx=None,
        sizes=args.size,
        batch_size=args.batch_size*4,
        shuffle=False,
        drop_last=False,
        num_workers=8,
    )

    # Initialize encoder and model
    encoder = Encoder(
        in_channels=x.shape[1],
        hidden_channels=args.hidden_channels,
        dropout=args.dropout,
        negative_slope=args.negative_slope,
    ).to(device)
    model = Model(
        encoder=encoder,
        decoder_hidden=args.decoder,
        project_hidden=args.projection,
        use_weight=args.use_weight,
        pos_weight_strategy=args.pos_weight_strategy,
        neg_weight_strategy=args.neg_weight_strategy,
    ).to(device)
    logging.info(f"Model architecture:\n{model}")

    optimizer = torch.optim.Adam(
        model.parameters(), lr=args.lr, weight_decay=args.weight_decay
    )
    
    # load_end = time.time()
    # logging.info(
    #     f"loading_time: {load_end-load_start:.2f}s"
    # )
    
    # Training loop
    train_start = time.time()
    step = 0
    
    use_epochs = args.epochs is not None and args.epochs > 0
    use_steps  = not use_epochs and args.max_steps is not None and args.max_steps > 0
    
    if use_epochs:
        for epoch in range(1, args.epochs + 1):
            model.train()
            for (bs, n_id, adjs), adj_batch, *rest in train_loader:
                # Move subgraph to device
                adjs = [adj.to(device) for adj in adjs]
                adj_batch = adj_batch.to(device)

                # Generate positive mask
                features = expr if args.embedding_type == "pheno_expr" else x
                pos_mask = get_positivePairs(adj_batch, features, strategy=args.strategy)

                optimizer.zero_grad()
                z, proj_z, rec_expr, _ = model(x[n_id], adjs=adjs)
                z = F.normalize(z, p=2, dim=1)
                proj_z = F.normalize(proj_z, p=2, dim=1)
                contrast_loss, recon_loss = model.compute_loss(
                    z, pos_mask, proj_z, rec_expr, target=expr
                )
                loss = 0.8 * contrast_loss + 0.2 * recon_loss if args.decoder else contrast_loss
                loss.backward()
                optimizer.step()

                step += 1
                if args.verbose:
                    logging.info(
                        f"Epoch {epoch} Step {step:04d} "
                        f"contrast_loss={contrast_loss:.4f}, recon_loss={recon_loss:.4f}"
                    )

    elif use_steps:
        while step < args.max_steps:
            model.train()
            for (bs, n_id, adjs), adj_batch, *rest in train_loader:
                # Move subgraph to device
                adjs = [adj.to(device) for adj in adjs]
                adj_batch = adj_batch.to(device)

                # Generate positive mask
                features = expr if args.embedding_type == "pheno_expr" else x
                pos_mask = get_positivePairs(adj_batch, features, strategy=args.strategy)

                optimizer.zero_grad()
                z, proj_z, rec_expr, _ = model(x[n_id], adjs=adjs)
                z = F.normalize(z, p=2, dim=1)
                proj_z = F.normalize(proj_z, p=2, dim=1)
                contrast_loss, recon_loss = model.compute_loss(
                    z, pos_mask, proj_z, rec_expr, target=expr
                )
                loss = 0.8 * contrast_loss + 0.2 * recon_loss if args.decoder else contrast_loss
                loss.backward()
                optimizer.step()

                step += 1
                if args.verbose:
                    logging.info(
                        f"Step {step:04d} "
                        f"contrast_loss={contrast_loss:.4f}, recon_loss={recon_loss:.4f}"
                    )
    else:
        logging.warning("Neither epochs nor max_steps specified, no training performed.")
    
    # logging.info(f"Training completed in {time.time() - train_start:.2f}s")

    # Inference: compute embeddings for all nodes
    model.eval()
    z_all = torch.zeros((x.shape[0], args.hidden_channels[-1]), device='cpu')
    with torch.no_grad():
        for (bs, n_id, adjs), _, batch in tqdm(test_loader, desc="Inference"):    
            
            adjs = [adj.to(device) for adj in adjs]
            z, _, _, _ = model(x[n_id].to(device), adjs=adjs)
            
            z_all[batch] = z.detach().cpu().float()

    z_all = F.normalize(z_all, p=2, dim=1)
    adata.obsm["CellNiche"] = z_all.numpy()

    # Clustering and metrics
    if args.metrics:
        adata, metrics = clustering_st(adata, n_classes, z_all, y, refine=args.refine)
        logging.info(f"Clustering metrics: {metrics}")
    # Save results
    if args.save:
        output_file = f"{args.save_path}/{args.dataset}_emb_{args.embedding_type}.h5ad"
        adata.write_h5ad(output_file)
        logging.info(f"Saved embeddings to {output_file}")



from typing import Callable, List, NamedTuple, Optional, Tuple, Union

import os
import pickle
import torch
from torch import Tensor
from torch_sparse import SparseTensor
from typing import Union, List, Optional, Callable
import numpy as np

from node2vec import Node2Vec
import networkx as nx


class EdgeIndex(NamedTuple):
    edge_index: Tensor
    e_id: Optional[Tensor]
    size: Tuple[int, int]

    def to(self, *args, **kwargs):
        edge_index = self.edge_index.to(*args, **kwargs)
        e_id = self.e_id.to(*args, **kwargs) if self.e_id is not None else None
        return EdgeIndex(edge_index, e_id, self.size)


class Adj(NamedTuple):
    adj_t: SparseTensor
    e_id: Optional[Tensor]
    size: Tuple[int, int]

    def to(self, *args, **kwargs):
        adj_t = self.adj_t.to(*args, **kwargs)
        e_id = self.e_id.to(*args, **kwargs) if self.e_id is not None else None
        return Adj(adj_t, e_id, self.size)


class NeighborSampler(torch.utils.data.DataLoader):
    r"""
    This code adapted from
    (https://github.com/pyg-team/pytorch_geometric/blob/master/torch_geometric/loader/neighbor_sampler.py).
    (https://github.com/EdisonLeeeee/MAGI/blob/master/magi/neighbor_sampler.py).
    """
    def __init__(self,
                 edge_index: Union[Tensor, SparseTensor],
                 adj: SparseTensor,
                 sizes: List[int],
                 is_train: bool = False,
                 wt: int = 20,
                 wl: int = 4,
                 biased_random_walk: bool = False,
                 p: float = 1.0,
                 q: float = 1.0,
                 drop_last= False,
                 node_idx: Optional[Tensor] = None,
                 num_nodes: Optional[int] = None,
                 return_e_id: bool = True,
                 transform: Callable = None,
                 sim_adj: SparseTensor = None,
                 **kwargs):

        edge_index = edge_index.to('cpu')

        if 'collate_fn' in kwargs:
            del kwargs['collate_fn']
        if 'dataset' in kwargs:
            del kwargs['dataset']

        # Save for Pytorch Lightning < 1.6:
        self.edge_index = edge_index
        self.adj = adj
        self.sim_adj = None
        self.node_idx = node_idx
        self.num_nodes = num_nodes
        self.is_train = is_train
        self.drop_last = drop_last
        self.wt = wt
        self.wl = wl
        self.biased_random_walk = biased_random_walk
        self.p = p
        self.q = q

        self.sizes = sizes
        self.return_e_id = return_e_id
        self.transform = transform
        self.is_sparse_tensor = isinstance(edge_index, SparseTensor)
        self.__val__ = None

        # Obtain a *transposed* `SparseTensor` instance.
        if not self.is_sparse_tensor:
            if (num_nodes is None and node_idx is not None
                    and node_idx.dtype == torch.bool):
                num_nodes = node_idx.size(0)
            if (num_nodes is None and node_idx is not None
                    and node_idx.dtype == torch.long):
                num_nodes = max(int(edge_index.max()), int(node_idx.max())) + 1
            if num_nodes is None:
                num_nodes = int(edge_index.max()) + 1

            value = torch.arange(edge_index.size(1)) if return_e_id else None
            self.adj_t = SparseTensor(row=edge_index[0], col=edge_index[1],
                                      value=value,
                                      sparse_sizes=(num_nodes, num_nodes)).t()
        else:
            adj_t = edge_index
            if return_e_id:
                self.__val__ = adj_t.storage.value()
                value = torch.arange(adj_t.nnz())
                adj_t = adj_t.set_value(value, layout='coo')
            self.adj_t = adj_t

        self.adj_t.storage.rowptr()

        if node_idx is None:
            node_idx = torch.arange(self.adj_t.sparse_size(0))
        elif node_idx.dtype == torch.bool:
            node_idx = node_idx.nonzero(as_tuple=False).view(-1)

        super().__init__(
            node_idx.view(-1).tolist(), collate_fn=self.sample, drop_last=self.drop_last, **kwargs)

    def _get_pos_neg_batch(self, random_nodes, biased_random_walk):
        random_nodes_count = random_nodes.shape[0]
        rowptr, col, _ = self.adj.csr()

        # -------------------------------
        # Stage 1: Full-graph random walk
        # -------------------------------
        random_nodes_repeat = random_nodes.repeat(self.wt)
        rw1 = self.adj.random_walk(random_nodes_repeat, self.wl)[:, 1:]
        # print(f"rw1.shape: {rw1.shape}")
        if not isinstance(rw1, torch.Tensor):
            rw1 = rw1[0]
        rw1 = rw1.t().reshape(-1, random_nodes_count).t()
        batch = []
        for i in range(random_nodes_count):
            rw_nodes, rw_times = torch.unique(rw1[i], return_counts=True)
            nodes = rw_nodes[rw_times > rw_times.float().mean()].tolist()
            batch += nodes
        batch += random_nodes.tolist()
        batch = torch.tensor(batch).unique()

        # -------------------------------
        # Stage 2: Biased/Unbiased random walk
        # -------------------------------
        batch_size = batch.shape[0]
        if biased_random_walk == True:
            walks_file = os.path.join("./results/walsk/", "node2vec_walks.pkl")
            if os.path.exists(walks_file):
                print(f"Found precomputed walks at {walks_file}, loading...")
                with open(walks_file, "rb") as f:
                    walks = pickle.load(f)
            else:
                print(f"No precomputed walks found. Computing full graph walks...")
                full_graph = self._create_graph_from_adj(self.adj)
                node2vec = Node2Vec(full_graph, walk_length=self.wl, num_walks=self.wt, p=self.p, q=self.q, workers=8)
                walks = node2vec.walks
                with open(walks_file, "wb") as f:
                    pickle.dump(walks, f)
                print(f"Walks saved to {walks_file}")

            node2vec_walks = [list(map(int, walk)) for walk in walks]
            rw2 = torch.tensor(node2vec_walks, dtype=torch.long)
            batch_set = set(batch.tolist())
            mask = []
            for i in range(rw2.size(0)):
                row_nodes = rw2[i].tolist()
                if row_nodes[0] in batch_set:
                    mask.append(True)
                else:
                    mask.append(False)
            mask = torch.tensor(mask, dtype=torch.bool)
            rw2 = rw2[mask]

        else:
            batch_repeat = batch.repeat(self.wt)
            rw2 = self.adj.random_walk(batch_repeat, self.wl)[:, 1:]
            if not isinstance(rw2, torch.Tensor):
                rw2 = rw2[0]

        rw2 = rw2.t().reshape(-1, batch_size).t()
        row, col, val = [], [], []
        for i in range(batch.shape[0]):
            rw2_nodes, rw2_times = torch.unique(rw2[i], return_counts=True)
            row += [batch[i].item()] * rw2_nodes.shape[0]
            col += rw2_nodes.tolist()
            val += rw2_times.tolist()

        unique_nodes = list(set(row + col))
        subg2g = dict(zip(unique_nodes, list(range(len(unique_nodes)))))

        row = [subg2g[x] for x in row]
        col = [subg2g[x] for x in col]
        idx = torch.tensor([subg2g[x] for x in batch.tolist()])

        adj_ = SparseTensor(row=torch.LongTensor(row), col=torch.LongTensor(col), value=torch.tensor(val),
                            sparse_sizes=(len(unique_nodes), len(unique_nodes)))

        adj_batch, _ = adj_.saint_subgraph(idx)
        adj_batch_sp = adj_batch.to_scipy(layout='coo')
        adj_batch_sp.setdiag([0] * idx.shape[0])
        adj_batch = SparseTensor.from_scipy(adj_batch_sp)

        if self.sim_adj is not None:
            sim_row, sim_col, sim_val = [], [], []
            for r, c in zip(row, col):
                sim_row.append(r)
                sim_col.append(c)

                adj_row, adj_col, adj_val = self.sim_adj.storage.row(), self.sim_adj.storage.col(), self.sim_adj.storage.value()
                mask = (adj_row == r) & (adj_col == c)
                if mask.any():
                    similarity = adj_val[mask].item()
                else:
                    similarity = 0.0 

                sim_val.append(similarity)

            sim_val_tensor = torch.tensor(sim_val, dtype=torch.float32)
            sim_adj_ = SparseTensor(row=torch.LongTensor(sim_row), col=torch.LongTensor(sim_col), value=sim_val_tensor,
                                    sparse_sizes=(len(unique_nodes), len(unique_nodes)))
            sim_adj_batch, _ = sim_adj_.saint_subgraph(idx)
            sim_adj_batch_sp = sim_adj_batch.to_scipy(layout='coo')
            sim_adj_batch_sp.setdiag([0] * idx.shape[0])
            sim_adj_batch = SparseTensor.from_scipy(sim_adj_batch_sp)

            return batch, adj_batch, sim_adj_batch

        return batch, adj_batch

    def sample(self, batch):
        if not isinstance(batch, Tensor):
            batch = torch.tensor(batch)
        adj_batch = None
        sim_adj_batch = None
        if self.sim_adj is not None:
            batch, adj_batch, sim_adj_batch = self._get_pos_neg_batch(batch, biased_random_walk=self.biased_random_walk)
        else:
            batch, adj_batch = self._get_pos_neg_batch(batch, biased_random_walk=self.biased_random_walk)
        batch_size: int = len(batch)

        adjs = []
        n_id = batch
        for size in self.sizes:
            adj_t, n_id = self.adj_t.sample_adj(n_id, size, replace=False)
            e_id = adj_t.storage.value()
            size = adj_t.sparse_sizes()[::-1]
            if self.__val__ is not None:
                adj_t.set_value_(self.__val__[e_id], layout='coo')

            if self.is_sparse_tensor:
                adjs.append(Adj(adj_t, e_id, size))
            else:
                row, col, _ = adj_t.coo()
                edge_index = torch.stack([col, row], dim=0)
                adjs.append(EdgeIndex(edge_index, e_id, size))

        adjs = adjs[0] if len(adjs) == 1 else adjs[::-1]
        out = (batch_size, n_id, adjs)
        out = self.transform(*out) if self.transform is not None else out

        if sim_adj_batch is not None:
            return out, adj_batch, sim_adj_batch, batch
        else:
            return out, adj_batch, batch

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(sizes={self.sizes})'



import os
import random
import numpy as np
import pandas as pd
import torch
import scipy
import logging
import torch.nn.functional as F
import scanpy as sc
from natsort import natsorted
from scipy.sparse import coo_matrix
from sklearn.preprocessing import LabelEncoder
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import KMeans
from sklearn.metrics import (
    adjusted_rand_score,
    normalized_mutual_info_score,
    adjusted_mutual_info_score,
    f1_score,
    silhouette_score,
)
from torch_sparse import SparseTensor
from torch_geometric.utils import to_undirected

from typing import Optional, Union, Any, Tuple, List

def load_data(
    data_path: str,
    dataset: str,
    phenoLabels: str,
    nicheLabels: Optional[str],
    embedding_type: str,
    radius: Optional[float],
    k_neighborhood: int,
    hvg: bool
) -> Tuple[torch.FloatTensor, torch.LongTensor, np.ndarray, Any, int, Optional[torch.FloatTensor]]:
    """
    Load AnnData from .h5ad, build node features, edge index, and labels.

    Args:
        data_path: Directory containing the .h5ad files.
        dataset: Filename (without extension) to load.
        phenoLabels: Column name in `adata.obs` for phenotype labels (one-hot).
        nicheLabels: Column name in `adata.obs` for true labels (or None to skip).
        embedding_type: One of 'pheno_expr', 'pheno', 'expr'.
        radius: Radius threshold (if not None) for radius graph.
        k_neighborhood: Number of neighbors for kNN graph if radius is None.
        hvg: Whether to select highly variable genes for expression.

    Returns:
        x: Node feature matrix (one-hot or expr) as FloatTensor.
        edge_index: LongTensor[2, E] of graph edges.
        y: True label array of shape [N].
        adata: The AnnData object.
        n_classes: Number of unique labels in y.
        expr: Expression matrix as FloatTensor if needed, else None.
    """
    # 1) Read AnnData
    path = os.path.join(data_path, f"{dataset}.h5ad")
    adata = sc.read_h5ad(path).copy()

    # 2) Build phenotype one-hot features
    pheno = adata.obs[phenoLabels].astype(str)
    ph_le = LabelEncoder().fit(pheno)
    ph_idx = ph_le.transform(pheno)
    onehot = torch.nn.functional.one_hot(
        torch.from_numpy(ph_idx), num_classes=len(ph_le.classes_)
    ).float()

    # 3) Build graph edge_index
    if "edgeList" in adata.uns:
        edge_np = np.array(adata.uns["edgeList"])
        edge_index = torch.from_numpy(edge_np).long()
        edge_index = to_undirected(edge_index)
    else:
        # choose coords
        if "spatial" in adata.obsm and adata.obsm["spatial"] is not None:
            coords = adata.obsm["spatial"]
            # coords = adata.obs[["x", "y"]].to_numpy() # spleen
        else:
            coords = adata.obs[["x", "y"]].to_numpy()

        if radius is not None:
            nbrs = NearestNeighbors(radius=radius).fit(coords)
            _, idxs = nbrs.radius_neighbors(coords)
            rows = np.concatenate([np.full(len(n), i) for i, n in enumerate(idxs)])
            cols = np.concatenate(idxs)
        else:
            nbrs = NearestNeighbors(n_neighbors=k_neighborhood+1).fit(coords)
            _, idxs = nbrs.kneighbors(coords)
            rows = np.repeat(np.arange(coords.shape[0]), k_neighborhood)
            cols = idxs[:, 1:].flatten()

        mat = coo_matrix((np.ones_like(rows), (rows, cols)),
                         shape=(coords.shape[0], coords.shape[0]))
        mat = mat + mat.T  # make undirected
        edge_index = torch.from_numpy(np.vstack(mat.nonzero()).astype(np.int64))
        
        neighbors_count = np.array([len(neighbors) for neighbors in idxs])
        average_neighbors = neighbors_count.mean()
        logging.info(f"Average number of neighbors per node: {average_neighbors}")
        # print(f"================ Average number of neighbors per node: {average_neighbors} ================")

    # 4) Encode true labels from nicheLabels if provided
    if nicheLabels is not None and nicheLabels in adata.obs:
        true_vals = adata.obs[nicheLabels].astype(str)
        nl_encoder = LabelEncoder().fit(true_vals)
        y = nl_encoder.transform(true_vals)
        n_classes = len(nl_encoder.classes_)
    else:
        # default dummy labels: all-zero, one class
        y = np.zeros(adata.n_obs, dtype=int)
        n_classes = 1

    # 5) Prepare expression matrix if needed
    expr: Optional[torch.FloatTensor] = None
    if embedding_type in ("pheno_expr", "expr"):
        if hvg:
            # select HVGs, normalize & log
            sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=512)
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)
            adata = adata[:, adata.var["highly_variable"]]

        mat = adata.X
        if scipy.sparse.isspmatrix(mat):
            arr = mat.toarray()
        else:
            arr = np.array(mat)
        expr = torch.from_numpy(arr).float()

    # logging.info(
    #     f"Loaded {dataset}: " f"{onehot.shape[0]} nodes, {edge_index.shape[1]} edges, {onehot.shape[0].shape[-1]} features"
    # )
    
    # 6) Return according to embedding type
    if embedding_type == "pheno_expr":
        logging.info(
            f"Loaded {dataset}: " f"{onehot.shape[0]} nodes, {edge_index.shape[1]} edges, {onehot.shape[-1]} features"
        )
        return torch.tensor(onehot, dtype=torch.float), edge_index, y, adata, n_classes, torch.tensor(expr, dtype=torch.float)
    elif embedding_type == "pheno":
        logging.info(
            f"Loaded {dataset}: " f"{onehot.shape[0]} nodes, {edge_index.shape[1]} edges, {onehot.shape[-1]} features"
        )
        return torch.tensor(onehot, dtype=torch.float), edge_index, y, adata, n_classes, None
    elif embedding_type == "expr":
        logging.info(
            f"Loaded {dataset}: " f"{expr.shape[0]} nodes, {edge_index.shape[1]} edges, {expr.shape[-1]} features"
        )
        return torch.tensor(expr, dtype=torch.float), edge_index, y, adata, n_classes, None
    else:
        raise ValueError(f"Unknown embedding_type: {embedding_type}")

def setup_seed(seed: int) -> None:
    """
    Set random seed for reproducibility across Python, NumPy, Torch, and CUDA.

    Args:
        seed (int): The seed to set.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    # torch.backends.cudnn.deterministic = True
    # torch.backends.cudnn.benchmark = False

def create_sparse_tensor_from_edges(
    rows: list[int],
    cols: list[int],
    sparse_size: tuple[int, int],
    device: torch.device = torch.device("cpu"),
) -> SparseTensor:
    """
    Create a SparseTensor from row/col indices.

    Args:
        rows (list[int]): Row indices.
        cols (list[int]): Column indices.
        sparse_size (tuple[int,int]): Matrix size.
        device (torch.device): Device for tensor.

    Returns:
        SparseTensor
    """
    vals = torch.ones(len(rows), device=device)
    return SparseTensor(
        row=torch.tensor(rows, device=device),
        col=torch.tensor(cols, device=device),
        value=vals,
        sparse_sizes=sparse_size,
    )

def sparse_intersection_and_union(
    adj1: SparseTensor,
    adj2: SparseTensor,
    strategy: str = "and",
) -> SparseTensor:
    """
    Compute intersection or union of two sparse adjacency matrices.

    Args:
        adj1, adj2 (SparseTensor): Input graphs.
        strategy (str): 'and' or 'or'.

    Returns:
        SparseTensor
    """
    rows1, cols1 = adj1.storage.row(), adj1.storage.col()
    rows2, cols2 = adj2.storage.row(), adj2.storage.col()
    device = rows1.device

    set1 = set(zip(rows1.tolist(), cols1.tolist()))
    set2 = set(zip(rows2.tolist(), cols2.tolist()))

    if strategy == "and":
        common = set1 & set2
    else:
        common = set1 | set2

    if not common:
        return SparseTensor(sparse_sizes=adj1.sparse_sizes())

    rows, cols = zip(*common)
    return create_sparse_tensor_from_edges(rows, cols, adj1.sparse_sizes(), device)

def get_positivePairs(
    subAdj: SparseTensor,
    features: Optional[torch.Tensor] = None,
    strategy: str = "freq",
) -> SparseTensor:
    """
    Generate positive pair adjacency based on strategy.

    Args:
        subAdj (SparseTensor): Stochastic subgraph adjacency.
        features (Tensor): Node features.
        strategy (str): 'freq', 'sim', 'and', 'or'.

    Returns:
        SparseTensor: positive adjacency.
    """
    row, col, val = subAdj.storage.row(), subAdj.storage.col(), subAdj.storage.value()
    # Frequency-based mask
    freq_thresh = subAdj.sum(dim=1) / subAdj.storage.colptr()[1:]
    mask = val > freq_thresh[row]
    rows, cols, vals = row[mask], col[mask], val[mask]
    freq_adj = SparseTensor(row=rows, col=cols, value=vals, sparse_sizes=subAdj.sparse_sizes())

    if strategy == "freq":
        return freq_adj

    if features is None:
        raise ValueError("Features required for non-freq strategy.")

    # Similarity-based mask
    f_row, f_col = features[row], features[col]
    sim_vals = F.cosine_similarity(f_row, f_col, dim=1)
    sim_vals[row == col] = 0
    sim_adj = SparseTensor(row=row, col=col, value=sim_vals, sparse_sizes=subAdj.sparse_sizes())
    sim_thresh = sim_adj.sum(dim=1) / sim_adj.storage.colptr()[1:]
    sim_mask = sim_vals > sim_thresh[row]
    sim_rows, sim_cols, sim_vals = row[sim_mask], col[sim_mask], sim_vals[sim_mask]
    sim_based = SparseTensor(row=sim_rows, col=sim_cols, value=sim_vals, sparse_sizes=subAdj.sparse_sizes())

    if strategy == "sim":
        return sim_based

    # AND / OR combination
    return sparse_intersection_and_union(freq_adj, sim_based, strategy)

def match_labels(true_labels, predicted_labels, n_classes):
    from scipy.optimize import linear_sum_assignment as linear_assignment

    cost_matrix = np.zeros((n_classes, n_classes))

    for i in range(n_classes):
        for j in range(n_classes):
            cost_matrix[i, j] = np.sum((true_labels == i) & (predicted_labels == j))

    row_ind, col_ind = linear_assignment(-cost_matrix)

    new_labels = np.copy(predicted_labels)
    for i, j in zip(row_ind, col_ind):
        new_labels[predicted_labels == j] = i
    return new_labels

def refine_spatial_domains(y_pred, coord, n_neighbors=6):
    nbrs = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(coord)
    distances, indices = nbrs.kneighbors(coord)
    indices = indices[:, 1:]

    y_refined = pd.Series(index=y_pred.index, dtype='object')

    for i in range(y_pred.shape[0]):
        y_pred_count = y_pred[indices[i, :]].value_counts()

        if y_pred[i] in y_pred_count.index:
            if (y_pred_count.loc[y_pred[i]] < n_neighbors / 2) and (y_pred_count.max() > n_neighbors / 2):
                y_refined[i] = y_pred_count.idxmax()
            else:
                # y_refined[i] = y_pred[i] # waring
                y_refined.iloc[i] = y_pred[i]
        else:
            y_refined.iloc[i] = y_pred[i]

    y_refined = pd.Categorical(
        values=y_refined.astype('U'),
        categories=natsorted(map(str, y_refined.unique())),
    )
    return y_refined

def clustering_st(
    adata: Any,
    n_clusters: int,
    features: Optional[Union[torch.Tensor, np.ndarray]] = None,
    true_labels: Optional[np.ndarray] = None,
    refine: bool = False,
) -> Tuple[Any, dict]:
    """
    Perform KMeans clustering and compute evaluation metrics.

    Args:
        adata (AnnData): Annotated data object.
        features (Tensor or ndarray): Embeddings.
        n_clusters (int): Number of clusters.
        true_labels (ndarray): Ground-truth labels.

    Returns:
        adata (AnnData): Updated with 'kmeans' clusters.
        metrics (dict): Cluster evaluation metrics.
    """
    # 1) Convert to numpy
    if torch.is_tensor(features):
        feats = features.cpu().numpy()
    else:
        feats = features
    # 2) Run KMeans
    km = KMeans(n_clusters=n_clusters, random_state=0)
    raw_labels = km.fit_predict(feats).astype(int)

    # 3) Store raw labels
    adata.obs['kmeans'] = pd.Categorical(raw_labels)
    clustering_results = {'kmeans': raw_labels}

    # 4) Optional spatial refinement
    if refine:
        # spatial coords must exist
        coords = adata.obsm.get('spatial')
        if coords is None:
            raise ValueError("adata.obsm['spatial'] needed for refinement")
        for method, labels in list(clustering_results.items()):
            refined = refine_spatial_domains(pd.Series(labels), coords)
            refined = refined.astype(int)
            col = f"{method}_refined"
            adata.obs[col] = pd.Categorical(refined)
            clustering_results[col] = refined

    # 5) Compute metrics
    metrics_results: dict = {}
    if true_labels is not None:
        for method, labels in clustering_results.items():
            # align predicted → true
            aligned = match_labels(true_labels, labels, n_clusters)

            acc = (aligned == true_labels).mean()
            nmi = normalized_mutual_info_score(true_labels, aligned)
            ari = adjusted_rand_score(true_labels, aligned)
            ami = adjusted_mutual_info_score(true_labels, aligned)
            f1m = f1_score(true_labels, aligned, average='macro')
            f1i = f1_score(true_labels, aligned, average='micro')
            sil = silhouette_score(feats, aligned)

            metrics_results[method] = {
                'Acc': acc,
                'NMI': nmi,
                'AMI': ami,
                'ARI': ari,
                'F1 Macro': f1m,
                'F1 Micro': f1i,
                'Silhouette': sil,
            }

    return adata, metrics_results    
    



from importlib import import_module
from types import ModuleType
from typing import Dict, Any

__version__ = "0.1.4"
__author__  = "ZMLiang <lzzzmmgpt@gmail.com>"

_EXPORTS = {
    # ── CLI (functions) ───────────────────────────────────────
    "main"          : (".main", "main"),
    "parse_args"    : (".main", "parse_args"),

    # ── Core model (classes) ─────────────────────────────────
    "Model"         : (".model", "Model"),
    "Encoder"       : (".model", "Encoder"),

    # ── Training (function + class) ──────────────────────────
    "run"           : (".trainer", "run"),
    "Trainer"       : (".trainer", "Trainer"),

    # ── Sampler (class) ──────────────────────────────────────
    "Sampler"       : (".sampler", "NeighborSampler"),

    # ── Utils (function) ─────────────────────────────────────
    "clustering_st" : (".utils", "clustering_st"),
}

_cache: Dict[str, Any] = {}

def __getattr__(name: str) -> Any:
    if name in _cache:
        return _cache[name]
    if name not in _EXPORTS:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    module_path, attr_name = _EXPORTS[name]
    module: ModuleType = import_module(module_path, __package__)
    obj = getattr(module, attr_name)
    _cache[name] = obj
    return obj

def __dir__() -> list[str]:
    return sorted(list(globals().keys()) + list(_EXPORTS.keys()))

__all__ = list(_EXPORTS.keys())


import argparse
import logging
import yaml

from .trainer import run

def parse_args(argv=None):
    parser = argparse.ArgumentParser(
        description="Entry point for CellNiche training & inference"
    )
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to YAML config file",
    )
    return parser.parse_args(argv)

def main(argv=None):
    # 1) Load config file
    args = parse_args(argv)
    with open(args.config, "r") as f:
        cfg = yaml.safe_load(f)

    # 2) Turn config keys into attributes of a simple args object
    class Obj: pass
    opts = Obj()
    for k, v in cfg.items():
        setattr(opts, k, v)

    # 3) Set up logging
    level = logging.INFO if opts.verbose else logging.WARNING
    logging.basicConfig(
        level=level,
        format="%(asctime)s %(levelname)s: %(message)s",
        datefmt="%H:%M:%S",
    )
    # 4) Kick off training/inference
    run(opts)

if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import BatchNorm1d, LeakyReLU, PReLU
from torch_geometric.nn import SAGEConv
from typing import Optional, List, Tuple



class ContrastiveLoss(nn.Module):
    """
    Contrastive loss for graph embeddings using a positive pair mask.

    Args:
        temperature (float): Initial temperature parameter for scaling similarities.
        scale_by_temperature (bool): If True, scales the loss by temperature.
        use_weight (bool): If True, applies weighting to positive pairs.
        pos_weight_strategy (str): Strategy for positive pair weights: 'none', 'inverse_freq', 'inverse_sim'.
        neg_weight_strategy (str): Strategy for negative pair weights: 'none', 'inverse_sim'.
        pos_generation_strategy (str): Strategy influencing positive weight generation.
    """
    def __init__(
        self,
        temperature: float = 0.5,
        scale_by_temperature: bool = False,
        use_weight: bool = True,
        pos_weight_strategy: str = 'inverse_sim',
        neg_weight_strategy: str = 'inverse_sim',
        pos_generation_strategy: str = 'inverse_sim',
    ):
        super().__init__()
        # Register temperature as a learnable parameter
        self.temperature = nn.Parameter(torch.tensor(temperature))
        self.scale_by_temperature = scale_by_temperature
        self.use_weight = use_weight
        self.pos_weight_strategy = pos_weight_strategy
        self.neg_weight_strategy = neg_weight_strategy
        self.pos_generation_strategy = pos_generation_strategy

    def forward(self, embeddings: torch.Tensor, pos_mask: torch.sparse.FloatTensor) -> torch.Tensor:
        """
        Compute contrastive loss.

        Args:
            embeddings (Tensor): Node embeddings of shape (batch, dim).
            pos_mask (SparseTensor): Sparse mask indicating positive pairs.

        Returns:
            Tensor: Scalar loss value.
        """
        device = embeddings.device
        batch_size = embeddings.size(0)

        # Normalize embeddings and scale by temperature
        z = F.normalize(embeddings, p=2, dim=1)
        temp = F.softplus(self.temperature)
        sim_matrix = torch.matmul(z, z.t()) / temp
        # For numerical stability
        sim_matrix = sim_matrix - sim_matrix.max(dim=1, keepdim=True).values.detach()

        # Mask out self-comparisons
        identity = torch.eye(batch_size, device=device).bool()

        # Construct dense positive and negative masks
        pos_dense = torch.zeros_like(sim_matrix, dtype=torch.bool)
        # pos_dense[pos_mask.indices()[0], pos_mask.indices()[1]] = True
        pos_dense[pos_mask.storage.row(), pos_mask.storage.col()] = True
        neg_dense = ~pos_dense & ~identity

        # Compute log probabilities
        log_probs = F.log_softmax(sim_matrix, dim=1)
        pos_log = log_probs[pos_dense]

        if self.use_weight:
            # Compute positive weights
            values = pos_mask.values().float()
            if self.pos_weight_strategy == 'inverse_freq':
                pos_w = 1.0 / (values + 1e-8)
            elif self.pos_weight_strategy == 'inverse_sim':
                pos_w = 1.0 - values
            else:
                pos_w = torch.ones_like(values)
            # Normalize weights to [0,1]
            pos_w = (pos_w - pos_w.min()) / (pos_w.max() - pos_w.min() + 1e-8)

            # Weighted positive loss
            loss = - (pos_log * pos_w).mean()
        else:
            # Standard contrastive loss
            loss = - pos_log.mean()

        return loss


class Encoder(nn.Module):
    """
    Graph Encoder with multiple SAGEConv layers.

    Args:
        in_channels (int): Dimensionality of input node features.
        hidden_channels (List[int]): Hidden sizes for each GNN layer.
        conv_layer (callable): Graph convolution class (default: SAGEConv).
        dropout (float): Dropout probability between layers.
        negative_slope (float): Negative slope for LeakyReLU.
    """
    def __init__(
        self,
        in_channels: int,
        hidden_channels: list[int],
        conv_layer=SAGEConv,
        dropout: float = 0.0,
        negative_slope: float = 0.5,
    ):
        super().__init__()
        # Build graph convolutional layers
        self.convs = nn.ModuleList(
            conv_layer(
                in_channels if i == 0 else hidden_channels[i - 1],
                hidden_channels[i]
            )
            for i in range(len(hidden_channels))
        )
        self.dropout = dropout
        self.negative_slope = negative_slope
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Reinitialize all convolutional layers."""
        for conv in self.convs:
            conv.reset_parameters()

    def forward(
        self,
        x: torch.Tensor,
        adjs: Optional[List[Tuple[torch.Tensor, Optional[torch.Tensor], Tuple[int,int]]]] = None,
        edge_index: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass through GNN layers.

        Args:
            x (Tensor): Node feature matrix.
            adjs (list, optional): List of (edge_index, e_id, size) for sampling.
            edge_index (Tensor, optional): Full-graph edge index if adjs is None.

        Returns:
            h (Tensor): Output node embeddings.
            target (Tensor): Features of target nodes for reconstruction.
        """
        if adjs is None:
            # Use full-graph edge index
            assert edge_index is not None, "edge_index required if adjs is None"
            layers = [(edge_index, None, (x.size(0), x.size(0)))]
            target = x
        else:
            layers = adjs
            # Target nodes correspond to first N' rows
            target = x[: layers[-1][2][1]]

        for i, (e_idx, _, size) in enumerate(layers):
            x = F.dropout(x, p=self.dropout, training=self.training)
            x_target = x[: size[1]]
            x = self.convs[i]((x, x_target), e_idx)
            x = F.leaky_relu(x, negative_slope=self.negative_slope)

        return x, target


class Model(nn.Module):
    """
    Full Graph Contrastive Model with optional projection and reconstruction.

    Args:
        encoder (Encoder): Graph encoder instance.
        decoder_hidden (List[int], optional): Hidden sizes for decoder MLP.
        project_hidden (List[int], optional): Hidden sizes for projection head.
        tau (float): Temperature parameter for contrastive loss.
        use_weight (bool): Use weighted contrastive loss.
        pos_weight_strategy (str): Positive weight strategy.
        neg_weight_strategy (str): Negative weight strategy.
    """
    def __init__(
        self,
        encoder: Encoder,
        decoder_hidden: Optional[List[int]] = None,
        project_hidden: Optional[List[int]] = None,
        tau: float = 0.5,
        use_weight: bool = True,
        pos_weight_strategy: str = 'none',
        neg_weight_strategy: str = 'none',
    ):
        super().__init__()
        self.encoder = encoder
        embed_dim = encoder.convs[-1].out_channels

        # Projection head for contrastive learning
        if project_hidden:
            layers = []
            in_dim = embed_dim
            for dim in project_hidden:
                layers.append(nn.Linear(in_dim, dim))
                layers.append(PReLU())
                in_dim = dim
            self.project = nn.Sequential(*layers)
        else:
            self.project = None

        # Decoder for reconstruction (optional)
        if decoder_hidden:
            layers = []
            in_dim = embed_dim
            for dim in decoder_hidden:
                layers.append(nn.Linear(in_dim, dim))
                layers.append(BatchNorm1d(dim))
                layers.append(LeakyReLU())
                in_dim = dim
            layers.append(nn.Linear(in_dim, encoder.convs[0].in_channels))
            self.decoder = nn.Sequential(*layers)
            self.recon_loss_fn = nn.MSELoss()
        else:
            self.decoder = None
            self.recon_loss_fn = None

        # Contrastive loss function
        self.loss_fn = ContrastiveLoss(
            temperature=tau,
            use_weight=use_weight,
            pos_weight_strategy=pos_weight_strategy,
            neg_weight_strategy=neg_weight_strategy,
        )

    def forward(
        self,
        x: torch.Tensor,
        adjs: Optional[List[Tuple[torch.Tensor, Optional[torch.Tensor], Tuple[int, int]]]] = None,
        edge_index: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Forward pass returns:
          - h: encoder output embeddings
          - z: projected embeddings (for contrastive loss)
          - rec: reconstructed features (for decoder)
          - target: original target features
        """
        h, target = self.encoder(x, adjs, edge_index)
        z = self.project(h) if self.project else h
        rec = self.decoder(h) if self.decoder else h
        
        return h, z, rec, target

    def compute_loss(
        self,
        z: torch.Tensor,
        pos_mask: torch.sparse.FloatTensor,
        proj_z: torch.Tensor,
        rec: Optional[torch.Tensor] = None,
        target: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Compute contrastive and reconstruction losses.

        Args:
            z (Tensor): Projected embeddings.
            pos_mask (SparseTensor): Mask of positive pairs.
            rec (Tensor, optional): Reconstructed features.
            target (Tensor, optional): Original features for reconstruction.

        Returns:
            contrast_loss, recon_loss (Tensors)
        """
        # Contrastive loss
        if self.project:
            contrast_loss = self.loss_fn(proj_z, pos_mask)
        else:
            contrast_loss = self.loss_fn(z, pos_mask)

        # Reconstruction loss
        if self.decoder:
            if target is None:
                raise ValueError("Target features required for reconstruction loss.")
            recon_loss = self.recon_loss_fn(rec, target)
        else:
            recon_loss = torch.tensor(0.0, device=z.device)

        return contrast_loss, recon_loss
